#!/usr/bin/python3

import sys, os, re, subprocess, configparser, time, errno, argparse, shutil, copy
import logging, logging.config
import serial, serial.tools.list_ports
from subprocess import Popen, PIPE, STDOUT

from enum import Enum

logging.config.fileConfig('/opt/hw-ppp/log.ini')
log = logging.getLogger('su.artix.hw.ppp')

class Error(Enum):
    '''
    Перечисление кастомных кодов возврата
    '''
    OK = 0                   # нет ошибки
    PPPD_ALREADY_STARTED = 3 # некоторые ppp-соединения уже запущены
    PPP_IFACE_DOWN = 4       # ppp-интерфейс пока не поднят
    IPTABLES_RULE_FAILS = 5  # ошибка создания правила iptables
    PPPD_START_FAIL = 6      # ошибка подключения ppp-интерфейса
    PORT_PREPARE_FAIL = 7    # ошибка подготовки порта
    SERVICE_DISABLE = 8      # сервис выключен
    CONNECTION_INTERRUPT = 9 # процесс подключения был прерван

def timer(timeout, period, func, *args):
    '''
    Простой блокирующий таймер
    Крутится в течение timeout секунд(или пока функция func не вернёт положительный результат)
    выполняя функцию func с параметрами *args в блокирующем режиме через каждые period секунд
    Возвращает True, если функция завершила работу успешно
    '''
    curTime = time.time()
    while time.time() - curTime < timeout:
        if func(*args):
            return True
        time.sleep(period)
    return False

class EnableServiceException(Exception):
    pass

class PppManager(object):
    '''
    Класс для управления ppp-соединениями сервиса hw-ppp
    Со сторонними ppp-соединениями сервис не работает
    '''
    # статические данные
    statusDir = '/opt/hw-ppp/status/'
    paramFilePath = '/opt/hw-ppp/hw-ppp.conf'
    pidFilePathOriginal = '/var/run/ppp-%s.pid'
    pidFilePath = statusDir + 'ppp-%s.pid'
    leftWidth = 40
    rightWidth = 20

    def __init__(self):
        '''
        При создании объекта проверить существование директории статус-файлов
        Если её нет, то создать
        '''
        self.oldStdout = sys.stdout
        self.devNull = open(os.devnull, 'w')
        self.defaultDeviceParams = {'vid': '', 'pid': '', 'hub': '', 'local_ip': '192.168.137.2', 'remote_ip': '192.168.137.3', 'network': '192.168.137.0', 'mask': '255.255.255.0'}

        config = configparser.ConfigParser()
        config.read(PppManager.paramFilePath)

        if not config.has_section('Main') or not config.getboolean('Main', 'enable'):
            raise EnableServiceException

        log.info("Попытка создать директорию '%s'" % PppManager.statusDir)
        try:
            os.mkdir(PppManager.statusDir)
        except OSError:
            log.info('Директория уже создана')

    def __del__(self):
        self.devNull.close()

    def start(self):
        log.info('Запуск соединений')
        if self.__checkAllStatuses():
            log.warning('Некоторые соединения ещё активны')
            print('Already started')
            return Error.PPPD_ALREADY_STARTED.value
        params = self.__readParams()
        result = Error.OK.value
        self.__enableIpV4Forward()
        for deviceNum in params:
            result = max(self.__connect(deviceNum, params), result)
        return result

    def stop(self):
        log.info('Остановка всех ppp соединений')
        for pidFile in [file for file in os.listdir(PppManager.statusDir) if file.endswith(".pid")]:
            log.info("Найден pid-файл '%s'" % (PppManager.statusDir + pidFile))
            self.__stopPpp(pidFile)
        self.__clearStatusDir()
        return Error.OK.value

    def status(self):
        result = self.__checkAllStatuses()
        if not result:
            print('All connections are closed')
        return (not result)

    def devicelist(self):
        for device in serial.tools.list_ports.comports(True):
            for hwidel in device.hwid.split(' '):
                result = re.search('VID:PID=([0-9|A-Z|a-z]{4})\:([0-9|A-Z|a-z]{4})', hwidel)
                if result:
                    print("- VID = '%s', PID = '%s', hub = '%s'" % (result.group(1), result.group(2), device.location))

    def __clearStatusDir(self):
        for file in os.listdir(PppManager.statusDir):
            filePath = os.path.join(PppManager.statusDir, file)
            try:
                if os.path.isfile(filePath):
                    os.remove(filePath)
            except Exception as e:
                log.info(str(e))

    def __checkAllStatuses(self):
        '''
        Проверка статусов всех ppp-процессов
        Возвращает:
            False - все соединения закрыты
            True - хотя бы одно соединение работает
        '''
        log.info('Получение статусов всех соединений')
        result = False
        for statusFile in [file for file in os.listdir(PppManager.statusDir) if file.endswith(".status")]:
            log.info("Найден файл-статус '%s'" % (PppManager.statusDir + statusFile))
            statusParams = self.__readStatusParams(PppManager.statusDir + statusFile)
            log.info("Параметры подключения: %s" % statusParams)
            linkname = statusParams.get('linkname')
            isProcessAlive = self.__statusPppProcess(linkname)
            result |= isProcessAlive
            if isProcessAlive:
                self.__printFormatString("Process '%s'" % linkname, '...exists')
                self.__statusPppIface(statusParams.get('ifname'))
            else:
                self.__printFormatString("Process '%s'" % linkname, "...doesn't exists")
        log.info('Статусы соединений получены')
        return result

    def __readParams(self):
        '''
        Чтение параметров из '/opt/hw-ppp/hw-ppp.conf'
        Возвращает мап [DEVICE.N][key][value]
        '''
        log.info("Чтение параметров из '%s'" % PppManager.paramFilePath)
        config = configparser.ConfigParser()
        config.read(PppManager.paramFilePath)

        params = {}
        for sectionName in config.sections():
            res = re.search('DEVICE.([0-9]*)$', sectionName)
            if not res:
                continue
            deviceNum = int(res.group(1))
            params[deviceNum] = copy.deepcopy(self.defaultDeviceParams)
            for k, v in config.items(sectionName):
                params[deviceNum][k] = v
            # если настройка пуста или отсутствует, то найти устройство по vid, pid, hub
            if not params[deviceNum].get('device'):
                params[deviceNum]['device'] = self.__findDevice(params[deviceNum].get('vid'), params[deviceNum].get('pid'), params[deviceNum].get('hub'))
        log.info("Параметры получены: '%s'" % str(params))
        return params

    def __findDevice(self, vid, pid, hub):
        log.info('Поиск устройства по VID(%s), PID(%s), hub(%s)' % (vid, pid, hub))
        for device in serial.tools.list_ports.comports(True):
            for hwidel in device.hwid.split(' '):
                result = re.search('VID:PID=([0-9|A-Z|a-z]{4})\:([0-9|A-Z|a-z]{4})', hwidel)
                if result and result.group(1) == vid and result.group(2) == pid and device.location == hub:
                    log.info("Найдено устройство '%s'" % device.device)
                    return device.device
        log.info('Устройство не найдено')
        return None

    def __connect(self, deviceNum, params):
        result = self.__connectPpp(deviceNum, params)
        if (result == Error.OK.value) or (result == Error.PPP_IFACE_DOWN.value):
            result = self.__createPostforwardRule(deviceNum, params)
        return result

    def __connectPpp(self, deviceNum, params):
        device = params[deviceNum].get('device')
        local_ip = params[deviceNum].get('local_ip')
        remote_ip = params[deviceNum].get('remote_ip')
        linkname = 'hw-%d' % deviceNum
        log.info("Создание подключения: устройство - '%s', локальный ip(хост) - '%s', удалённый ip(ФР) - '%s'" % (device, local_ip, remote_ip))
        args = [
            'pppd',
            'call',
            'hw-ppp',
            'linkname',
            linkname,
            'connect',
            'chat -v -f /opt/hw-ppp/winclient.chat',
            'lock',
            'passive',
            '%s:%s' % (local_ip, remote_ip),
            device,
            '115200'
        ]
        # старт демона pppd
        proc = subprocess.Popen(args, stdout=PIPE, stderr=STDOUT)
        sys.stdout.write("Starting on '%s', local ip = %s, remote ip = %s" % (device, local_ip, remote_ip))
        sys.stdout.flush()
        result = Error.OK.value
        # проверка того, что демон стартанул
        proc.poll()
        if proc.returncode:
            log.error('Ошибка при выполении: pppd вернул код ошибки %d' % proc.returncode)
            print('   ...failed')
            result = Error.PPPD_START_FAIL.value
        else:
            log.info('Process pid = %s' % str(proc.pid))
            self.__saveOriginalPppPid(linkname, proc.pid)
            sys.stdout = self.devNull
            resultPppIface = False
            try:
                resultPppIface = self.__waitForIfaceCreated(60, linkname)
            except KeyboardInterrupt:
                log.warn('Процесс подключения был прерван')
                self.stop()
                result = Error.CONNECTION_INTERRUPT.value
            sys.stdout = self.oldStdout
            print('   ...Ok') if resultPppIface else print('   ...failed')
        return result

    def __createPostforwardRule(self, deviceNum, params):
        log.info('Создание правила iptables')
        network = params[deviceNum].get('network')
        mask = params[deviceNum].get('mask')
        args = [
            'iptables',
            '-t',
            'nat',
            '-D',
            'POSTROUTING',
            '-s',
            network + '/' + mask,
            '-d',
            '0/0',
            '-j',
            'MASQUERADE'
        ]
        # Сначала удаляем правило. Ошибки удаления правила не интересны
        self.__callCmd(args)
        # Теперь добавляем правило. Здесь ошибки добавления уже нужно обработать
        args[3] = '-I'
        callResult = self.__callCmd(args)
        self.__printFormatString('Creating iptables forward rule', '...' + ('failed' if callResult else 'Ok'))
        if callResult:
            log.error("Ошибка создания правила + '%s'" % callResult)
            return Error.IPTABLES_RULE_FAILS.value
        log.info('Правило создано')
        return Error.OK.value

    def __enableIpV4Forward(self):
        log.info('Активация форвардинга IPv4')
        args = [
            'sysctl',
            '-w',
            'net.ipv4.ip_forward=1'
        ]
        callResult = self.__callCmd(args)
        self.__printFormatString("Activate IPv4 forwarding", '...' + ('failed' if callResult else 'Ok'))
        if callResult:
            log.error('Ошибка активации форвардинга')
        else:
            log.info('Форвардинг активирован')

    def __waitForIfaceCreated(self, timeout, linkname):
        return timer(timeout, 2, self.__statusPppIfaceByLinkname, linkname)

    def __saveOriginalPppPid(self, linkname, pid):
        with open(PppManager.pidFilePath % linkname, 'w') as pidFile:
            pidFile.write(str(pid))

    def __statusPppIfaceByLinkname(self, linkname):
        params = self.__getStatusParams(linkname)
        if not params:
            return False
        return self.__statusPppIface(params.get('ifname'))

    def __readStatusParams(self, statusFile):
        '''
        Чиение параметров из 'statusFile'
        Возвращает мап [key][value] из секции 'Properties'
        '''
        config = configparser.ConfigParser()
        config.read(statusFile)

        params = {}
        try:
            for k, v in config.items('Properties'):
                params[k] = v
        except configparser.Error as e:
            pass
        return params

    def __callCmd(self, cmd):
        '''
        Вызов процесса 'cmd' и ожидание его завершения
        Возвращает retcode выполненного процесса
        '''
        cmd = [str(arg) for arg in cmd]
        log.info("Вызов '%s'" % ' '.join(cmd))
        proc = subprocess.Popen(cmd, stdout=PIPE, stderr=PIPE)
        result = proc.communicate()
        if result[1]:
            log.warn("Вызываемая команда вернула ошибку: '%s'" % result[1])
        else:
            log.info('Успешно')
        return proc.returncode

    def __checkProcessExistance(self, pid):
        '''
        Проверка существования процесса по 'pid'
        '''
        if not pid:
            log.warning('Некорректное значение pid')
            return False
        try:
            os.kill(pid, 0)
        except OSError as err:
            if err.errno == errno.ESRCH:
                log.warning('Процесс с pid(%d) не существует' % pid)
                return False
        log.info('Процесс с pid(%d) найден' % pid)
        return True

    def __getStatusFilePath(self, linkname):
        '''
        Возвращает путь до статус файла по имени процесса ppp-соединения
        '''
        return PppManager.statusDir + linkname + '.status'


    def __getStatusParams(self, linkname):
        '''
        Возвращает параметры из статус-файла по имени процесса ppp-соединения
        '''
        return self.__readStatusParams(self.__getStatusFilePath(linkname))

    def __stopPpp(self, pidFilePath):
        '''
        Остановка pppd процесса
        '''
        pid = None
        try:
            with open(PppManager.statusDir + pidFilePath, 'r') as pidFile:
                pid = int(pidFile.readline().strip())
            if pid:
                os.kill(pid, 9)
            else:
                log.warning('pid не обнаружен')
        except Exception as e:
            log.error(str(e))
        log.info("Процесс с pid '%s' остановлен" % str(pid))

    def __statusPppIface(self, ifname):
        '''
        Проверка наличия ppp-интерфейса
        Если интерфейс не обнаружен или его имя не задано, то возвращает False
        '''
        log.info("Получение статуса интерфейса '%s'" % ifname)
        result = False
        if ifname:
            callResult = self.__callCmd(['ip', 'link', 'show', ifname])
            width = 30 - len(ifname)
            if callResult:
                log.warning('Интерфейс не доступен')
            else:
                log.info('Интерфейс поднят')
            self.__printFormatString(ifname, '...' + ('not worked' if callResult else 'established'))
            result = not callResult
        else:
            log.error("Не задан параметр 'ifname'")
            result = False
        log.info('Статус интерфейса получен')
        return result

    def __statusPppProcess(self, linkname):
        '''
        Проверка наличия процесса pppd
        Если процесс запущен, то возвращает True
        '''
        log.info("Получение статуса процесса с именем pid '%s'" % linkname)
        result = False
        if linkname:
            pid = None
            try:
                with open(PppManager.pidFilePath % linkname, 'r') as pidFile:
                    pid = int(pidFile.readline().strip())
                result = self.__checkProcessExistance(pid)
            except IOError as e:
                log.error(str(e))
                result = False
        else:
            log.error("Не задан параметр 'linkname'")
            result = False
        log.info('Статус процесса получен')
        return result

    def __printFormatString(self, left, right):
        print((('{:<%d}' % PppManager.leftWidth).format(left) + ('{:<%d}' % PppManager.rightWidth).format(right)))

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Ppp connection manager for some fiscal registers')
    parser.add_argument('CONTROL', choices=['start', 'stop', 'status', 'devicelist'])
    args = parser.parse_args()

    try:
        manager = PppManager()
    except EnableServiceException:
        log.info('Сервис выключен')
        exit(Error.SERVICE_DISABLE.value)

    log.info("Запуск скрипта с параметром '%s'" % args.CONTROL)
    exit(getattr(manager, args.CONTROL)())
